from .flatwrapper import FlatWrapper


class CompareFlatWrapper(FlatWrapper):
    def transform_node(self):
        # TODO : node transformation here
        self.original_node.twin.left = self.builder.ast_load(
            self.original_node.left.temp_name
        )

        comparators = []

        for node in self.original_node.comparators:
            comparators.append(self.builder.ast_load(node.temp_name))
        # TODO : node transformation here
        self.original_node.twin.comparators = comparators

    def flatten_children(self):
        left = self.original_node.left
        self.flattened_children.extend(left.wrapper.flat_wrap())

        for node in self.original_node.comparators:
            self.flattened_children.extend(node.wrapper.flat_wrap())



import copy
import ast
import inspect
from inspect import isfunction, ismethod, isclass, ismodule


# must adapt for Python 3 in the future
def remove_decorators(func):
    closure = func.__closure__
    if closure:
        for cell in closure:
            contents = cell.cell_contents
            if contents is func:
                continue

            if isfunction(contents) or ismethod(contents) or isclass(contents):
                new_func = remove_decorators(contents)
                if new_func:
                    return new_func
        else:
            return func
    else:
        return func


class MetaLink(object):
    def __init__(self, metaobject, selector, control="before", arguments=None):
        self.metaobject = metaobject
        self.selector = selector
        self.control = control
        self.nodes = []
        self.arguments = arguments or []
        self.reified_arguments = []
        self.option_arg_as_array = False
        self.literal_value_reifications = []

    def add_node(self, rf_node):
        self.nodes.append(rf_node)

    def remove_node(self, rf_node):
        self.nodes.remove(rf_node)
        rf_node.links.remove(self)

    def uninstall(self, from_node=None):
        nodes = [from_node] if from_node else list(self.nodes)
        for node in nodes:
            self.remove_node(node)
            node.method_node.reflective_method.restore()

    def reset_reified_arguments(self):
        self.reified_arguments = []


class ReflectiveMethod(object):
    def __init__(self, class_or_object, method, method_name):
        self.target_entity = class_or_object
        self.method_name = method_name
        self.original_method = method
        self.original_ast = None
        self.reflective_ast = None
        self.link_registry = dict()
        self.init_reflective_method()
        self.literal_value_reifications = dict()

    def init_reflective_method(self):
        builder = AstBuilder()
        self.original_ast = builder.rf_ast_for_method(
            self.target_entity, self.original_method, self.method_name
        )

        self.reflective_ast = copy.deepcopy(self.original_ast)
        self.original_ast.reflective_method = self

    def lookup_link(self, metalink_id):
        return self.link_registry[metalink_id]

    def add_link(self, metalink):
        self.link_registry[id(metalink)] = metalink

    def find_node_of_id_in_link(self, node_id, metalink):
        return next(node for node in metalink.nodes if node.rf_id == node_id)

    def add_literal_value_reification(self, value):
        self.literal_value_reifications[id(value)] = value

    def get_literal_value_reifications(self, value_id):
        return self.literal_value_reifications[value_id]

    def compile_rf_method(self, rf_ast, method_name):
        locs = {}
        compiled_method = compile(rf_ast, "<ast>", "exec")
        global_vars = {
            "__rf_original_method__": self.original_method,
            "__rf_method__": self,
        }
        global_vars.update(self.original_method.__globals__)

        eval(compiled_method, global_vars, locs)
        if not (isclass(self.target_entity) or ismodule(self.target_entity)):
            method = locs[method_name].__get__(self.target_entity)
        else:
            method = locs[method_name]

        setattr(self.target_entity, method_name, method)

    def recompile(self):
        self.compile_rf_method(self.reflective_ast, self.method_name)

    def invalidate(self):
        self.reflective_ast.body[0].body = self.original_ast.wrapper.flat_wrap()
        ast.fix_missing_locations(self.reflective_ast)
        self.recompile()

    def restore(self):
        self.link_registry = dict()
        self.init_reflective_method()
        self.recompile()


class AstBuilder(object):
    def __init__(self):
        self.method_node = None
        self.current_index = 1
        self.flattened_nodes = {}

    def ast_load_list(self, args_list):
        return ast.List(elts=args_list, ctx=ast.Load())

    def assign_named_value(self, name, value):
        value_store = self.ast_store(name)
        return self.ast_assign(value_store, value)

    def ast_assign(self, store_node, value_node):
        return ast.Assign(targets=[store_node], value=value_node)

    def ast_store(self, var_name):
        return ast.Name(id=var_name, ctx=ast.Store())

    def ast_load(self, var_name):
        load = ast.Name(id=var_name, ctx=ast.Load())
        load.temp_name = var_name
        return load

    def ast_expr(self, node):
        return ast.Expr(node)

    @staticmethod
    def get_method_source(method):
        method = remove_decorators(method)
        while "__rf_original_method__" in method.__globals__:
            method = method.__globals__["__rf_original_method__"]
        lines = inspect.getsourcelines(method)
        first_line = lines[0][0]
        indent = len(first_line) - len(first_line.lstrip())
        src = ""
        for line in lines[0]:
            src += line[indent:]
        return src

    @classmethod
    def ast_for_method(cls, method):
        return ast.parse(cls.get_method_source(method))

    def rf_ast_for_method(self, method_class, method, method_name):
        method_node = self.ast_for_method(method)
        method_node.method_name = method_name
        method_node.method_class = method_class
        return self.build_rf_ast(method_node)

    def build_rf_ast(self, node):
        self.method_node = node
        self.method_node.is_method = True
        self.visit_node(self.method_node)
        self.visit_clear_twins(node)
        self.visit_twins(copy.deepcopy(node))
        return node

    # We clear the twins before copying nodes,
    # to avoid recursions during deep copies.
    # Twins are set back in visit_twins
    def visit_clear_twins(self, node):
        for node in ast.iter_child_nodes(node):
            original_node = self.flattened_nodes[node.rf_id]
            original_node.twin = None
            self.visit_twins(node)

    def visit_twins(self, copy_node):
        for node in ast.iter_child_nodes(copy_node):
            original_node = self.flattened_nodes[node.rf_id]
            node.method_node = original_node.method_node
            original_node.twin = node
            self.visit_twins(node)

    def visit_node(self, node):
        self.decorate_node(node)

        for child in ast.iter_child_nodes(node):
            child.parent = node
            node.children.append(child)
            self.visit_node(child)

        return node

    def build_rf_node(self, node):
        self.decorate_node(node)
        return self.decorate_node(node)

    def decorate_node(self, node):
        if node != self.method_node:
            node.is_method = False

        node.is_generated = False

        node.rf_id = self.current_index
        self.flattened_nodes[node.rf_id] = node
        self.current_index += 1

        node_class = node.__class__
        node.can_be_wrapped = True
        node.temp_name = "temp_{}_{}".format(node_class.__name__, node.rf_id)

        node.method_node = self.method_node

        node.method_class = self.method_node.method_class
        node.links = list()

        from .wrappers import flat_wrappers

        if node_class in flat_wrappers:
            node.wrapper = flat_wrappers[node_class](node)
        else:
            node.wrapper = flat_wrappers["generic"](node)

        node.children = []
        return node



from .flatwrapper import FlatWrapper


class CFlowConditionExtractor(object):
    def __init__(self):
        self.flattened_condition = []
        self.preambles = []
        self.body_supplements = []

    def visit_node(self, rf_node):
        visit_method = "visit_" + rf_node.__class__.__name__
        getattr(self, visit_method)(rf_node)

    def visit_While(self, rf_node):
        test = rf_node.test
        if not test.wrapper.should_wrap(test):
            return
        self.extract_test(test)
        self.preambles = self.flattened_condition
        self.body_supplements = self.flattened_condition

    def visit_For(self, rf_node):
        pass

    def visit_If(self, rf_node):
        test = rf_node.test
        if not test.wrapper.should_wrap(test):
            return
        self.extract_test(test)
        self.preambles = self.flattened_condition

    def extract_test(self, rf_test):
        self.flattened_condition = rf_test.wrapper.flat_wrap()


class CFlowFlatWrapper(FlatWrapper):
    def __init__(self, rf_node):
        super(CFlowFlatWrapper, self).__init__(rf_node)

    def extract_body(self):
        return self.original_node.body

    def transform_node(self):
        # TODO : node transformation here
        self.original_node.twin.body = self.transform_body()
        self.extract_condition()
        self.node_transformation.append(self.original_node.twin)

    def should_wrap(self, rf_node):
        return True

    def transform_body(self):
        transformations = []
        for node in self.extract_body():
            if hasattr(node, "can_be_wrapped"):
                body_transformation = node.wrapper.flat_wrap()
                transformations.extend(body_transformation)
        return transformations

    def extract_condition(self):
        extractor = CFlowConditionExtractor()
        extractor.visit_node(self.original_node)
        self.flattened_children.extend(extractor.preambles)
        self.original_node.twin.body.extend(extractor.body_supplements)



import ast
from .flatwrapper import FlatWrapper
from .assign_flatwrapper import AssignFlatWrapper
from .return_flatwrapper import ReturnFlatWrapper
from .method_flatwrapper import MethodFlatWrapper
from .cflow_flatwrapper import CFlowFlatWrapper
from .compare_flatwrapper import CompareFlatWrapper
from .literal_flatwrapper import LiteralFlatWrapper
from .expr_flatwrapper import ExprFlatWrapper
from .call_flatwrapper import CallFlatWrapper


flat_wrappers = {
    ast.Assign: AssignFlatWrapper,
    ast.Return: ReturnFlatWrapper,
    ast.Module: MethodFlatWrapper,
    ast.If: CFlowFlatWrapper,
    ast.While: CFlowFlatWrapper,
    ast.For: CFlowFlatWrapper,
    ast.Compare: CompareFlatWrapper,
    ast.Num: LiteralFlatWrapper,
    ast.Str: LiteralFlatWrapper,
    ast.Name: LiteralFlatWrapper,
    ast.Expr: ExprFlatWrapper,
    ast.Call: CallFlatWrapper,
    "generic": FlatWrapper,
}


__all__ = [
    "AssignFlatWrapper",
    "ReturnFlatWrapper",
    "MethodFlatWrapper",
    "CFlowFlatWrapper",
    "CompareFlatWrapper",
    "LiteralFlatWrapper",
    "ExprFlatWrapper",
    "CallFlatWrapper",
    "FlatWrapper",
]



import ast
from .flatwrapper import FlatWrapper


class ReturnFlatWrapper(FlatWrapper):
    def __init__(self, rf_node):
        super(ReturnFlatWrapper, self).__init__(rf_node)
        self.additional_after_links = []

    def transform_node(self):
        original_node = self.original_node
        builder = self.builder
        value_assign = builder.assign_named_value(
            original_node.temp_name, original_node.value
        )
        self.node_transformation.append(builder.build_rf_node(value_assign))

    def basic_wrap(self):
        super(ReturnFlatWrapper, self).basic_wrap()
        self.append_links(self.additional_after_links)
        builder = self.builder
        return_node = ast.Return(builder.ast_load(self.original_node.temp_name))
        self.body.append(builder.build_rf_node(return_node))
        return self.body

    def should_wrap(self, rf_node):
        return (
            super(ReturnFlatWrapper, self).should_wrap(rf_node)
            or self.additional_after_links
        )



import ast
from ..core import AstBuilder
from ..reifications import ReificationGenerator, link_reification


class FlatWrapper(object):
    def __init__(self, rf_node):
        self.body = []
        self.preambles = []
        self.flattened_children = []
        self.before_links = []
        self.after_links = []
        self.instead_links = []
        self.original_node = rf_node
        self.reifications = set()
        self.node_transformation = []
        self.builder = AstBuilder()
        self.builder.method_node = self.original_node.method_node
        self.reifyer = ReificationGenerator()

    def reset_wrapping(self):
        self.body = []
        self.preambles = []
        self.flattened_children = []
        self.before_links = []
        self.after_links = []
        self.instead_links = []
        self.node_transformation = []

    def gen_preambles(self):
        preambles = self.reifyer.generate_reifications(self.original_node)
        self.preambles.extend(preambles)

    def flatten_children(self):
        pass

    def transform_node(self):
        self.node_transformation.append(self.original_node)

    def gen_link_node(self, link):
        arguments = []
        arguments.extend(link.reified_arguments)

        metaobject_node = ast.Attribute(
            value=link_reification(link), attr="metaobject", ctx=ast.Load()
        )
        attr_node = ast.Attribute(
            value=metaobject_node, attr=link.selector, ctx=ast.Load()
        )
        call_node = ast.Call(func=attr_node, args=arguments, keywords=[])

        return ast.Expr(call_node)

    def sort_links(self):
        for link in self.original_node.links:
            if link.control == "before":
                self.before_links.append(link)
            if link.control == "after":
                self.after_links.append(link)
            if link.control == "instead":
                self.instead_links.append(link)

    def flat_wrap(self):
        self.reset_wrapping()
        self.gen_preambles()
        self.sort_links()

        if not self.should_wrap(self.original_node):
            self.body.append(self.original_node)
            return self.body

        self.init_wrapping()
        return self.basic_wrap()

    def init_wrapping(self):
        self.flatten_children()
        self.transform_node()

    def basic_wrap(self):
        self.append_flattened_children()
        self.append_preambles()
        self.append_links(self.before_links)
        self.append_node_transformation()
        self.append_links(self.after_links)
        return self.body

    def should_wrap(self, rf_node):
        return self.should_wrap_node(rf_node) or self.should_wrap_children(rf_node)

    def should_wrap_node(self, rf_node):
        if rf_node.links:
            return True
        return False

    def should_wrap_children(self, rf_node):
        for node in rf_node.children:
            if self.should_wrap(node):
                return True
        return False

    def append_preambles(self):
        for node in self.preambles:
            self.body.append(node)

    def append_flattened_children(self):
        for node in self.flattened_children:
            if hasattr(node, "wrapper"):
                self.body.extend(node.wrapper.flat_wrap())
            else:
                self.body.append(node)

    def append_node_transformation(self):
        if self.instead_links:
            self.body.append(self.gen_link_node(self.instead_links[0]))
            return

        for node in self.node_transformation:
            self.body.append(node)

    def append_links(self, links):
        for link in links:
            self.body.append(self.gen_link_node(link))



import ast
from .flatwrapper import FlatWrapper


class AssignFlatWrapper(FlatWrapper):
    def flatten_children(self):
        value_node = self.original_node.value

        if self.is_node_call_with_links(value_node):
            self.flattened_children.append(value_node)
            return

        value_assign = self.builder.assign_named_value(
            self.original_node.temp_name, value_node
        )
        rf_assign = self.builder.build_rf_node(value_assign)
        self.flattened_children.append(rf_assign)

    def transform_node(self):
        if self.original_node.is_generated:
            self.node_transformation.append(self.original_node)
            return

        targets = self.original_node.targets
        value_node = self.original_node.value
        temp_name = self.original_node.temp_name
        if self.is_node_call_with_links(value_node):
            temp_name = value_node.temp_name
        value = self.builder.ast_load(temp_name)
        transformed_assign = ast.Assign(targets=targets, value=value)
        self.node_transformation.append(transformed_assign)

    def is_node_call_with_links(self, rf_node):
        if rf_node.__class__ is not ast.Call:
            return False
        return rf_node.wrapper.should_wrap(rf_node)



from .flatwrapper import FlatWrapper


class CallFlatWrapper(FlatWrapper):
    def flatten_children(self):
        self.flattened_children.extend(self.original_node.args)
        self.flattened_children.append(self.extract_receiver_node())

    def transform_node(self):
        twin = self.original_node.twin
        new_args = []
        for arg in self.original_node.args:
            new_args.append(self.builder.ast_load(arg.temp_name))

        twin.args = new_args
        twin.func.value = self.builder.ast_load(twin.func.value.temp_name)
        transformed_node = self.builder.assign_named_value(
            self.original_node.temp_name, twin
        )
        self.node_transformation.append(transformed_node)

    def extract_receiver_node(self):
        receiver = self.original_node.func.value

        if receiver.links:
            return receiver

        return self.builder.assign_named_value(receiver.temp_name, receiver)



from .flatwrapper import FlatWrapper


class LiteralFlatWrapper(FlatWrapper):
    def transform_node(self):
        assign = self.builder.assign_named_value(
            self.original_node.temp_name, self.original_node
        )
        self.node_transformation.append(self.builder.build_rf_node(assign))

    def should_wrap(self, rf_node):
        return True



"""
Reflectivity implementation for Python base package
"""
from .reflectivity import (
    link,
    reflective_ast_for_method,
    reflective_method_for,
    uninstall_all,
    metalinks,
    rf_methods,
)
from .core import MetaLink, ReflectiveMethod

__version__ = "0.1.0"

__all__ = [
    "link",
    "reflective_ast_for_method",
    "reflective_method_for",
    "uninstall_all",
    "metalinks",
    "rf_methods",
    "MetaLink",
    "ReflectiveMethod",
]



import ast
from .core import AstBuilder

builder = AstBuilder()


class Reification(object):
    def visit_node(self, rf_node, metalink):
        return self.visit_method(rf_node)(rf_node, metalink)

    def visit_method(self, rf_node):
        visit_method = "visit_" + rf_node.__class__.__name__
        return getattr(self, visit_method)


class LiteralValueReification(Reification):
    def __init__(self, value):
        self.value = value

    def visit_node(self, rf_node, metalink):
        rf_node.method_node.reflective_method.add_literal_value_reification(self.value)
        attr_node = ast.Attribute(
            value=rf_method_reification(),
            attr="get_literal_value_reifications",
            ctx=ast.Load(),
        )
        return ast.Call(func=attr_node, args=[ast.Num(id(self.value))], keywords=[])


class LinkReification(Reification):
    def visit_node(self, rf_node, metalink):
        return link_reification(metalink)


class ClassReification(Reification):
    def visit_node(self, rf_node, metalink):
        return ast.Attribute(
            value=node_reification(), attr="method_class", ctx=ast.Load()
        )


class ObjectReification(Reification):
    def visit_node(self, rf_node, metalink):
        return ast.Name(id="self", ctx=ast.Load())


class NodeReification(Reification):
    def visit_node(self, rf_node, metalink):
        attr_node = ast.Attribute(
            value=rf_method_reification(),
            attr="find_node_of_id_in_link",
            ctx=ast.Load(),
        )
        return ast.Call(
            func=attr_node,
            args=[ast.Num(rf_node.rf_id), link_reification(metalink)],
            keywords=[],
        )


class MethodReification(Reification):
    def visit_node(self, rf_node, metalink):
        return original_method_reification()


class SenderReification(Reification):
    def visit_node(self, rf_node, metalink):
        method_node_node = ast.Attribute(
            value=node_reification(), attr="method_node", ctx=ast.Load()
        )
        return ast.Attribute(value=method_node_node, attr="method_name", ctx=ast.Load())


class ReceiverReification(Reification):
    def visit_node(self, rf_node, metalink):
        return ast.Name(id=rf_node.func.value.temp_name, ctx=ast.Load())


class SelectorReification(Reification):
    def visit_node(self, rf_node, metalink):
        return ast.Str(rf_node.func.attr)


class ValueReification(Reification):
    def visit_Assign(self, assign_node, metalink):
        return ast.Name(id=assign_node.targets[0].id, ctx=ast.Load())

    def visit_Name(self, name, metalink):
        return name


class OldValueReification(ValueReification):
    def visit_Assign(self, assign_node, metalink):
        return ast.Name(id=assign_node.targets[0].id, ctx=ast.Load())


class NewValueReification(ValueReification):
    def visit_Assign(self, assign_node, metalink):
        return ast.Name(id=assign_node.temp_name, ctx=ast.Load())


class NameReification(Reification):
    def visit_Assign(self, assign_node, metalink):
        return self.visit_Name(assign_node.targets[0], metalink)

    def visit_Name(self, name_node, metalink):
        return ast.Str(name_node.id)


class ArgumentReification(Reification):
    def visit_Module(self, rf_node, metalink):
        return self.visit_FunctionDef(rf_node.body[0], metalink)

    def visit_FunctionDef(self, rf_node, metalink):
        args = []
        for arg in rf_node.args.args:
            if not arg.arg == "self":
                args.append(ast.Name(id=arg.arg, ctx=ast.Load()))

        return ast.List(elts=args, ctx=ast.Load())

    def visit_Call(self, rf_node, metalink):
        args = []
        for arg in rf_node.args:
            args.append(ast.Name(id=arg.id, ctx=ast.Load()))

        return ast.List(elts=args, ctx=ast.Load())


reifications_dict = {
    "class": ClassReification,
    "node": NodeReification,
    "object": ObjectReification,
    "method": MethodReification,
    "sender": SenderReification,
    "receiver": ReceiverReification,
    "selector": SelectorReification,
    "name": NameReification,
    "value": ValueReification,
    "old_value": OldValueReification,
    "new_value": NewValueReification,
    "arguments": ArgumentReification,
    "link": LinkReification,
}


def reification_for(key, metalink):
    if key in reifications_dict:
        return reifications_dict[key]()
    return LiteralValueReification(key)


def rf_method_reification():
    return builder.ast_load("__rf_method__")


def link_reification(link):
    link_id_node = ast.Num(id(link))

    link_attr_node = ast.Attribute(
        value=rf_method_reification(), attr="lookup_link", ctx=ast.Load()
    )
    return ast.Call(func=link_attr_node, args=[link_id_node], keywords=[])


def node_reification():
    return ast.Attribute(
        value=rf_method_reification(), attr="reflective_ast", ctx=ast.Load()
    )


def original_method_reification():
    return ast.Attribute(
        value=rf_method_reification(), attr="original_method", ctx=ast.Load()
    )


class ReificationGenerator(object):
    def __init__(self):
        self.reification_counter = 0
        self.arg_list = []

    def generate_reifications(self, rf_node):
        expressions = []

        for link in rf_node.links:
            link.reset_reified_arguments()

            for arg in link.arguments:
                reification = reification_for(arg, link).visit_node(rf_node, link)
                rf_name = self.rf_name_for_arg(arg, str(rf_node.rf_id))
                expressions.append(builder.assign_named_value(rf_name, reification))
                self.add_reified_argument_to_link(builder.ast_load(rf_name), link)

            if link.option_arg_as_array:
                link.reified_arguments.append(builder.ast_load_list(self.arg_list))
                self.arg_list = []

        return expressions

    def add_reified_argument_to_link(self, arg_node, metalink):
        if metalink.option_arg_as_array:
            self.arg_list.append(arg_node)
        else:
            metalink.reified_arguments.append(arg_node)

    def rf_name_for_arg(self, arg, rf_id):
        if isinstance(arg, str):
            return arg + "_" + rf_id

        self.reification_counter += 1
        return "value_{}_{}".format(rf_id, self.reification_counter)



import ast
from .cflow_flatwrapper import CFlowFlatWrapper


class MethodFlatWrapper(CFlowFlatWrapper):
    def transform_node(self):
        self.node_transformation.extend(self.transform_body())

    def sort_links(self):
        super(MethodFlatWrapper, self).sort_links()

        body = self.extract_body()
        if not body:
            return

        for rf_node in body:
            self.inject_after_links_in_method_returns(rf_node)

        if self.is_return(self.last_node()):
            self.after_links = []

    def is_return(self, rf_node):
        return rf_node.__class__ is ast.Return

    def extract_body(self):
        return self.original_node.body[0].body

    def last_node(self):
        body = self.extract_body()
        if body:
            return body[-1]
        return None

    def inject_after_links_in_method_returns(self, rf_node):
        if self.is_return(rf_node):
            rf_node.wrapper.additional_after_links.extend(self.after_links)
            return

        for child in rf_node.children:
            self.inject_after_links_in_method_returns(child)



from weakref import WeakValueDictionary, WeakSet
from .core import ReflectiveMethod

rf_methods = WeakValueDictionary()

metalinks = WeakSet()
nodes_with_links = WeakSet()


def reflective_ast_for_method(class_or_object, method_name):
    """
    Finds the reflective AST for the given method name, or generates it if it
    does not exist
    """
    return reflective_method_for(class_or_object, method_name).original_ast


def reflective_method_for(class_or_object, method_name):
    rf_key = (class_or_object, method_name)
    try:
        return rf_methods[rf_key]
    except KeyError:
        method = getattr(class_or_object, method_name)
        rf_method = ReflectiveMethod(class_or_object, method, method_name)
        rf_methods[rf_key] = rf_method
        return rf_method


def link(metalink, rf_node):
    metalinks.add(metalink)
    nodes_with_links.add(rf_node)

    metalink.add_node(rf_node)
    rf_node.links.append(metalink)
    rf_method = rf_node.method_node.reflective_method
    rf_method.add_link(metalink)
    rf_method.invalidate()


def uninstall_all():
    for metalink in metalinks:
        for node in metalink.nodes:
            metalink.remove_node(node)

    for node in nodes_with_links:
        node.links.clear()

    metalinks.clear()
    nodes_with_links.clear()

    for rf_method in rf_methods.values():
        rf_method.restore()



from .flatwrapper import FlatWrapper


class ExprFlatWrapper(FlatWrapper):
    def flat_wrap(self):
        self.reset_wrapping()
        if self.should_wrap_children(self.original_node):
            self.body.extend(self.original_node.value.wrapper.flat_wrap())
        else:
            self.body.append(self.original_node)
        return self.body



#!/usr/bin/env python

from setuptools import setup
import reflectivipy


packages = ["reflectivipy", "reflectivipy.wrappers"]

setup(
    name="reflectivipy",
    version=reflectivipy.__version__,
    description=(
        "A Python Implementation of the Reflectivity API from " "the Pharo language"
    ),
    long_description=open("README.rst").read(),
    keywords="object-centric partial-behavior-reflection metaprogramming",
    url="https://github.com/StevenCostiou/reflectivipy",
    author="Steven Costiou",
    author_email="steven.costiou@abc.fr",
    packages=packages,
    package_data={"": ["README.rst", "LICENCE"]},
    include_package_data=True,
    tests_require=["pytest"],
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Programming Language :: Python",
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: Implementation :: PyPy",
        "Operating System :: OS Independent",
        "Intended Audience :: Developers",
        "Topic :: Software Development",
        "Topic :: Software Development :: Libraries",
        "Topic :: Software Development :: Code Generators",
        "Topic :: Software Development :: Debuggers",
    ],
)
